// Created by Damien Katz and placed into the public domain. Enjoy.

/*
* This is the main file for the Agent Boost project. This file ontains the exports necessary
* to be invoked by the Notes Extension Manager.
*/


#include "Windows.h"
#include "Shlwapi.h"

#include <string>
#include <set>
#include <algorithm>
#include <map>
#include <stdio.h>
#include <new> // to makes the new operator throw if it failed allocation

#pragma pack(push, 1) // Set the packing in the following Notes Headers to 1 byte alignment
	#include "global.h"
	#include "addin.h"
	#include "nsfdb.h"
	#include "nsfnote.h"
	#include "ostime.h"
	#include "osmisc.h"
	#include "osmem.h"
	#include "extmgr.h"
	#include "nsferr.h"
	#include "oserr.h"
	#include "miscerr.h"
	#include "nif.h"
	#include "agents.h"
	#include "osenv.h"
	#include "kfm.h"
#pragma pack(pop)	// Reset packing to normal

#include "AgentBoost.h"
#include "NotesWrapper.h"
#include "Utilities.h"
#include "CacheObjects.h"

// ************************** namespace imports ******************************

using namespace std;
using namespace NotesWrapper;


// ************************** Macros ******************************

#define DEBUGOUT if (gDebugOut) DebugOut



// ************************** Type declarations ******************************

struct EventData
{
	STATUS		currentStatus;	// The result status of the event that just occured. 
								// For pre-events it is always NOERROR.
	STATUS		newStatus;	// The status to return back to the caller. If this is a
							// pre-event anything other ERR_EM_CONTINUE

	NOTEID		noteId;	// The note-id of the event note, if applicable.
	UNID		noteUnid;	// The note-UNID of the event note, if applicable.
	WORD		eventId;	// The ID of the event that is occuring
	bool		isPreEvent;	// true is pre event, false if post event
	NOTEHANDLE	hNote;	// The handle to the event note, if applicable
	DBHANDLE	hDb;	// The handle to the event database, if applicable
	DWORD		updateFlags;	// The update flags of the event, if applicable
	DWORD		openFlags;	// The open flags of the event, if applicable
	WORD		noteClass;	// The note class of the note, if unknown, the class is set to NOTE_CLASS_ALL
	bool		noteClassLookupFailed; // true if we tried to find the class information and failed
	AUTO_NOTEHANDLE	hNewNote;	// The new note to return back to the caller on NoteOpen events.
								// See the LSSetReturnNote() function for more info
};



// ************************** Function declarations ******************************

void RunAgent(DBHANDLE hDb, const string& agentName, NOTEHANDLE hNote);
void EnsureCache(bool firstTime);
void AddToCache(string& dbFilename, string& agentName, int eventId, bool beforeEvent, bool global, WORD noteClassMask, string& recordId);
string GetTextFromSummaryBufferItem(BYTE* pValue, WORD itemLen);
STATUS LNCALLBACK ExtMgrCallback (EMRECORD far *pExtRec);
int LoadCacheEntries(HCOLLECTION hCol, char* pSystemName);
bool IsMainProcess();
bool MatchesClass(const Booster* pBooster, EventData* pEventData);
void __cdecl ThrowExceptionNewHandler();

// **************************  Global variables ******************************

// Process wide cache of config DB
ContainerToFilenameMap gCacheDbMap;
BoosterContainer gCacheGlobal;

// Process wide mutex protecting cache
Mutex gCacheMutex;

// The extension manager registration handles for all the events
AUTO_HEMREGISTRATION ghEmRegNoteUpdateBefore;
AUTO_HEMREGISTRATION ghEmRegNoteUpdateAfter;

AUTO_HEMREGISTRATION ghEmRegNoteUpdateExtBefore;
AUTO_HEMREGISTRATION ghEmRegNoteUpdateExtAfter;

AUTO_HEMREGISTRATION ghEmRegNoteUpdateMailboxBefore;
AUTO_HEMREGISTRATION ghEmRegNoteUpdateMailboxAfter;

AUTO_HEMREGISTRATION ghEmRegNoteDeleteBefore;
AUTO_HEMREGISTRATION ghEmRegNoteDeleteAfter;

AUTO_HEMREGISTRATION ghEmRegNoteDeleteExtBefore;
AUTO_HEMREGISTRATION ghEmRegNoteDeleteExtAfter;

AUTO_HEMREGISTRATION ghEmRegNoteOpenBefore;
AUTO_HEMREGISTRATION ghEmRegNoteOpenAfter;

AUTO_HEMREGISTRATION ghEmRegNoteOpenExtBefore;
AUTO_HEMREGISTRATION ghEmRegNoteOpenExtAfter;

AUTO_HEMREGISTRATION ghEmRegNoteOpenByUnidBefore;
AUTO_HEMREGISTRATION ghEmRegNoteOpenByUnidAfter;

// Location and ID of config Db
char gConfigDbPath[MAXPATH+1];
DBID gConfigDbId;

// ID of log.nsf Db
DBID gLogDbId;

// Last time the config was updated, used during cache refreshing
TIMEDATE gConfigLastModified;

// Last time the config update was checked.
DWORD gLastRefreshTimeTickCount = 0;

// frequency to check config db for modifications
DWORD gEnsureCacheMs = DEFAULT_CACHE_REFRESH_MS;

// A place to store event data during an event.
ThreadLocalStorage eventTls;

bool gDebugOut = false; // flag to print debugging messages to console

bool gIsMainProcess = true; // true if we're the main process, else we're a secondary process

bool gInitialized = false; // have we been through our initialization in MainEntryPoint?


// **************************  The main exports ******************************

/*
* This the proc that get invoked once per start-up of each Notes/Domino Process.
* Each process must register the appropriate callback functions.
*/
STATUS LNPUBLIC MainEntryPoint (void)
{
	STATUS err = NOERROR;

	if (gInitialized) {
		return err;
	}

	gInitialized = true;

	try {
		WORD recursionID;
		CHAR buffer[MAXSPRINTF];

		set_new_handler(ThrowExceptionNewHandler); // set up so new operator throw exceptions on failure.

		int agentBoostDebug = OSGetEnvironmentInt(DEBUG_ENV_VAR);

		gDebugOut = (agentBoostDebug != 0);

		gIsMainProcess = IsMainProcess();

		// Log a simple message to the console. This is so we can always tell
		// if it loaded successfully

		if (gIsMainProcess) {
			// but only if its the main process. Otherwise we'll get this message
			// for every single server process. yuck!
			ConsoleOut(VERSION_STRING " initialized.");
		}

		DEBUGOUT("Agent Boost Loading");

		if (!OSGetEnvironmentString(CONFIGDB_ENV_VAR, gConfigDbPath, sizeof(gConfigDbPath)-1)) {
			strcpy(gConfigDbPath, DEFAULT_CONFIG_DB);
		}

		if (OSGetEnvironmentString(CACHE_REFRESH_MS_ENV_VAR, buffer, sizeof(buffer)-1)) {
			gEnsureCacheMs = atoi(buffer);
		}

		AUTO_DBHANDLE hConfigDb;
		AUTO_DBHANDLE hLogDb;

		CheckStatus(
			NSFDbOpen(gConfigDbPath, &hConfigDb),
			"Couldn't open %s.", gConfigDbPath);

		CheckStatus(
			NSFDbIDGet(hConfigDb, &gConfigDbId),
			"Huh? Couldn't get DBID for config database!");

		memset(&gLogDbId, 0, sizeof(gLogDbId));

		if (NOERROR == NSFDbOpen("log.nsf", &hLogDb)) {
			NSFDbIDGet(hConfigDb, &gLogDbId);
		}

		TimeConstant(TIMEDATE_MINIMUM, &gConfigLastModified);

		EnsureCache(true);

		// Register the recursion id and all the events.

		CheckStatus(
			EMCreateRecursionID(&recursionID), 
			"Couldn't create recursion id."
			);

		CheckStatus(
			EMRegister(EM_NSFNOTEUPDATEMAILBOX, EM_REG_BEFORE, ExtMgrCallback, recursionID, &ghEmRegNoteUpdateMailboxBefore)
			);

		CheckStatus(
			EMRegister(EM_NSFNOTEUPDATEMAILBOX, EM_REG_AFTER, ExtMgrCallback, recursionID, &ghEmRegNoteUpdateMailboxAfter)
			);

		CheckStatus(
			EMRegister(EM_NSFNOTEUPDATE, EM_REG_BEFORE, ExtMgrCallback, recursionID, &ghEmRegNoteUpdateBefore)
			);

		CheckStatus(
			EMRegister(EM_NSFNOTEUPDATE, EM_REG_AFTER, ExtMgrCallback, recursionID, &ghEmRegNoteUpdateAfter)
			);

		CheckStatus(
			EMRegister(EM_NSFNOTEUPDATEXTENDED, EM_REG_BEFORE, ExtMgrCallback, recursionID, &ghEmRegNoteUpdateExtBefore)
			);

		CheckStatus(
			EMRegister(EM_NSFNOTEUPDATEXTENDED, EM_REG_AFTER, ExtMgrCallback, recursionID, &ghEmRegNoteUpdateExtAfter)
			);

		CheckStatus(
			EMRegister(EM_NSFNOTEDELETE, EM_REG_BEFORE, ExtMgrCallback, recursionID, &ghEmRegNoteDeleteBefore)
			);

		CheckStatus(
			EMRegister(EM_NSFNOTEDELETE, EM_REG_AFTER, ExtMgrCallback, recursionID, &ghEmRegNoteDeleteAfter)
			);

		CheckStatus(
			EMRegister(EM_NSFNOTEOPENEXTENDED, EM_REG_BEFORE, ExtMgrCallback, recursionID, &ghEmRegNoteOpenExtBefore)
			);

		CheckStatus(
			EMRegister(EM_NSFNOTEOPENEXTENDED, EM_REG_AFTER, ExtMgrCallback, recursionID, &ghEmRegNoteOpenExtAfter)
			);

		CheckStatus(
			EMRegister(EM_NSFNOTEOPEN, EM_REG_BEFORE, ExtMgrCallback, recursionID, &ghEmRegNoteOpenBefore)
			);

		CheckStatus(
			EMRegister(EM_NSFNOTEOPEN, EM_REG_AFTER, ExtMgrCallback, recursionID, &ghEmRegNoteOpenAfter)
			);

		CheckStatus(
			EMRegister(EM_NSFNOTEOPENBYUNID, EM_REG_BEFORE, ExtMgrCallback, recursionID, &ghEmRegNoteOpenByUnidBefore)
			);

		CheckStatus(
			EMRegister(EM_NSFNOTEOPENBYUNID, EM_REG_AFTER, ExtMgrCallback, recursionID, &ghEmRegNoteOpenByUnidAfter)
			);

	}
	catch (notes_error& error) {
		ConsoleOut(error.what());
		err = error.status;
	}
	catch (exception& error) {
		ConsoleOut(error.what());
		err = ERR_UNKNOWN;
	}

	return err;
}

/*
* This is the Extension Manager event callback proc. It gets called whenever
* a registered event occurs.
*
* The status that is returned determines the outcome of the event. If 
*/
STATUS LNCALLBACK ExtMgrCallback (EMRECORD* pExtRec)
{
	STATUS err = NOERROR;
	VARARG_PTR varptr = pExtRec->Ap;
	NOTEHANDLE* phReturnNote = NULL; // used by EM_NSFNOTEOPENEXTENDED events

	try {

		vector<Booster> dbBoosters;
		vector<Booster> globalBoosters;
		vector<Booster>::const_iterator pBooster;

		// First thing is to extract out all the information
		// about the event and put in in the eventData structure.

		// init the EventData structure
		EventData eventData;

		memset(&eventData, 0, sizeof(eventData));

		eventData.currentStatus = pExtRec->Status;
		eventData.newStatus = ERR_EM_CONTINUE; // by default we continue the event.
		eventData.eventId = pExtRec->EId;
		eventData.isPreEvent = (pExtRec->NotificationType == EM_BEFORE);

		eventData.noteClass = NOTE_CLASS_ALL;	// this is a bit field. The bit that is set determines
												// the note class for this event. By default set all
												// bits (NOTE_CLASS_ALL) and assume this event applies
												// all note classes. If we have a note we will reset it.

		// Set the thread local storage variable to our eventData.
		// This is so any invoked agents can "callback" into us and query
		// for the event data. See LSGetEventData() for more information.
		eventTls.Set(&eventData);

		// Dump in the relevant info from the event
		switch (eventData.eventId)
		{
		case EM_NSFNOTEUPDATE:

			eventData.hNote = VARARG_GET(varptr, NOTEHANDLE);
			eventData.updateFlags = VARARG_GET(varptr, WORD);

			if (eventData.updateFlags & UPDATE_DELETED) {
				// make this event look like EM_NSFNOTEDELETE
				eventData.eventId = EM_NSFNOTEDELETE;
			}
			else {
				// make this event look like EM_NSFNOTEUPDATEXTENDED
				eventData.eventId = EM_NSFNOTEUPDATEXTENDED;
			}
			break;

		case EM_NSFNOTEUPDATEXTENDED:
		case EM_NSFNOTEUPDATEMAILBOX:
			eventData.hNote = VARARG_GET(varptr, NOTEHANDLE);
			eventData.updateFlags = VARARG_GET(varptr, DWORD);
			
			if (eventData.updateFlags & UPDATE_DELETED) {
				// make this event look like EM_NSFNOTEDELETE
				eventData.eventId = EM_NSFNOTEDELETE;
			}
			break;

		case EM_NSFNOTEDELETE:
			eventData.hDb = VARARG_GET(varptr, DBHANDLE);
			eventData.noteId = VARARG_GET(varptr, NOTEID);
			eventData.updateFlags = VARARG_GET(varptr, WORD);
			break;

		case EM_NSFNOTEOPEN:
			// make this event look like EM_NSFNOTEOPENEXTENDED
			eventData.eventId = EM_NSFNOTEOPENEXTENDED;

			eventData.hDb = VARARG_GET(varptr, DBHANDLE);
			eventData.noteId = VARARG_GET(varptr, NOTEID);
			eventData.openFlags =  VARARG_GET(varptr, WORD);
			phReturnNote = VARARG_GET(varptr, HANDLE*);			
			if (!eventData.isPreEvent && phReturnNote && eventData.currentStatus == NOERROR) {
				eventData.hNote = *phReturnNote;
			}

			break;

		case EM_NSFNOTEOPENEXTENDED:
			eventData.hDb = VARARG_GET(varptr, DBHANDLE);
			eventData.noteId = VARARG_GET(varptr, NOTEID);
			eventData.openFlags =  VARARG_GET(varptr, DWORD);
			VARARG_GET(varptr, DWORD); // skip
			VARARG_GET(varptr, BYTE*); // skip
			phReturnNote = VARARG_GET(varptr, HANDLE*);
			if (!eventData.isPreEvent && phReturnNote && eventData.currentStatus == NOERROR) {
				eventData.hNote = *phReturnNote;
			}
			break;

		case EM_NSFNOTEOPENBYUNID:
			// make this event look like EM_NSFNOTEOPENEXTENDED
			eventData.eventId = EM_NSFNOTEOPENEXTENDED;

			eventData.hDb = VARARG_GET(varptr, DBHANDLE);
			UNID* pUnid = VARARG_GET(varptr, UNID*);
			if (pUnid) {
				eventData.noteUnid = *pUnid;
			}
			eventData.openFlags =  VARARG_GET(varptr, WORD);
			phReturnNote = VARARG_GET(varptr, HANDLE*);
			if (!eventData.isPreEvent && phReturnNote && eventData.currentStatus == NOERROR) {
				eventData.hNote = *phReturnNote;
			}
			break;
		}

		if (eventData.hNote) {
			// Get various things about the hNote
			NSFNoteGetInfo(eventData.hNote, _NOTE_CLASS, &eventData.noteClass);
			ORIGINATORID oid;
			NSFNoteGetInfo(eventData.hNote, _NOTE_OID, &oid);
			eventData.noteUnid.File = oid.File;
			eventData.noteUnid.Note = oid.Note;

			// if we don't have a db handle already, get one from the hnote
			if (eventData.hDb == NULLHANDLE) {
				NSFNoteGetInfo(eventData.hNote,_NOTE_DB, &eventData.hDb);
			}
		}

		char pathBuffer[MAXPATH];
		string path;
		if (eventData.hDb) {
			// Never, ever respond to an event in the config database.
			// Otherwise if the user writes a bad agent they might have no way to
			// turn it off. The code below makes that check.
			//
			// Also, responding to events in log.nsf is also bad, since they can cause an
			// infinite loop if the agent logs anything.

			DBID dbId;
			
			CheckStatus(
				NSFDbIDGet(eventData.hDb, &dbId),
				"Couldn't get DBID for event database!");

			if (memcmp(&gConfigDbId, &dbId, sizeof(dbId)) == 0) {
				goto done; // its the config db, ABORT
			}
			if (memcmp(&gLogDbId, &dbId, sizeof(dbId)) == 0) {
				goto done; // its the log db, ABORT
			}

			CheckStatus(
				NSFDbPathGet(eventData.hDb, pathBuffer, NULL),
				"Couldn't get path for event database!");
		}
		else {
			pathBuffer[0] = 0;
		}

		path = pathBuffer;

		// Now get all the agents that are supposed to respond to this event
		{
			// locked
			MutexLock lock(gCacheMutex);

			EnsureCache(false);

			if (eventData.hDb) {
				BoosterContainer* pContainer = gCacheDbMap.GetContainer(path);

				if (pContainer) {
					dbBoosters = pContainer->GetBoosters(eventData.eventId, eventData.isPreEvent);
				}
			}

			globalBoosters = gCacheGlobal.GetBoosters(eventData.eventId, eventData.isPreEvent);

			// unlocked
		}
		

		// Now run the db specific agents
		for (pBooster = dbBoosters.begin(); pBooster != dbBoosters.end(); pBooster++) {
			if (MatchesClass(&(*pBooster), &eventData)) {
				try {
					RunAgent(eventData.hDb, pBooster->agentName, eventData.hNote);
				}
				catch (exception& e) {
					ConsoleOut(e.what());
				}

				if (eventData.newStatus != ERR_EM_CONTINUE) {
					break;
				}
			}
		}
		
		// now run the glogal agents
		for (pBooster = globalBoosters.begin(); pBooster != globalBoosters.end(); pBooster++) {
			if (pBooster->noteClassMask & eventData.noteClass) {
				AUTO_DBHANDLE hAgentDb;
				try {
					CheckStatus(
						NSFDbOpen(pBooster->dbName.c_str(), &hAgentDb),
						"Couldn't open database %s for agent %s", pBooster->dbName.c_str(), pBooster->agentName.c_str()
						);

					RunAgent(hAgentDb, pBooster->agentName, eventData.hNote);
				}
				catch (exception& e) {
					ConsoleOut(e.what());
				}

				if (eventData.newStatus != ERR_EM_CONTINUE) {
					break;
				}
			}
		}

		switch (eventData.eventId)
		{
		case EM_NSFNOTEUPDATE:
		case EM_NSFNOTEUPDATEXTENDED:
		case EM_NSFNOTEUPDATEMAILBOX:
		case EM_NSFNOTEDELETE:
			// no clean up necessary
			break;

		case EM_NSFNOTEOPENEXTENDED:
			if (eventData.newStatus != ERR_EM_CONTINUE && eventData.newStatus != NOERROR) {
				// if we are returning an error, then we should close the note 
				// because the caller will not free the returned hNote if there is an error
				if (!eventData.isPreEvent && *phReturnNote && eventData.hNote) {
					NSFNoteClose(eventData.hNote);
					*phReturnNote = NULLHANDLE; // Sometimes callers ignore the status code
												// and look to see if their handle was
												// populated, so unpopulate it
				}
			}
			else if (phReturnNote && eventData.hNewNote) {
				// we are returning a note an agent loaded up
				if (eventData.hNote) {
					// if there is already a note loaded then free it.
					NSFNoteClose(eventData.hNote);
				}

				*phReturnNote = eventData.hNewNote.Release();

				if (eventData.isPreEvent) {
					// This is occuring before NSF is supposed to load the note
					// but we already loaded the note, so indicate that
					// the work is already done
					eventData.newStatus = NOERROR;
				}

			}
			break;
		}

done:

		err = eventData.newStatus;
	}
	catch (exception& e)
	{
		ConsoleOut(e.what());
	}

	// Clear out the thread local event data variable, it's no longer valid.
	eventTls.Set(NULL);

	return err;
}

// **************************  Helper functions ******************************

/*
* Indicates if the Booster and the Event match Note-class wise.
*/
bool MatchesClass(const Booster* pBooster, EventData* pEventData)
{
	if (pBooster->noteClassMask == NOTE_CLASS_ALL) {
		// always a match!
		return true;
	}

	if (pEventData->noteClass == NOTE_CLASS_ALL) {

		// we don't have a class for the event yet
		// maybe we can lookup one up.

		if (!pEventData->noteClassLookupFailed) {
			// this means the type lookup hasn't been performed yet. Lets do it
			
			if (pEventData->hDb == NULL) {
				// oops, we have no database
				pEventData->noteClassLookupFailed = true;
			}
			else {
				WORD noteClass = 0;
				UNID blankUnid = {0,0,0,0};

				if (pEventData->noteId) {
					// we have a note id, use that
					if (NSFDbGetNoteInfo(
											pEventData->hDb,
											pEventData->noteId,
											NULL,	NULL,
											&noteClass)) {
						pEventData->noteClassLookupFailed = true;
					}
				
				}
				else if (memcmp(&pEventData->noteUnid, &blankUnid, sizeof(UNID)) != 0) {
					// we have a unid, use that
					if (NSFDbGetNoteInfoByUNID(
												pEventData->hDb,
												&pEventData->noteUnid,
												NULL, NULL,	NULL,
												&noteClass)) {

						pEventData->noteClassLookupFailed = true;
					}
				}
				else {					
					// we have no unid or note id, we can't look it up.
					pEventData->noteClassLookupFailed = true;
				}

				if (!pEventData->noteClassLookupFailed) {
					pEventData->noteClass = noteClass; // it worked, now lets set our event value
				}
			}
		}
	}

	if (pEventData->noteClass == NOTE_CLASS_ALL) {
		// The note class of the event is unknown.
		// This Booster is set to run only against certain documents classes. Since
		// this documents class can't be determined then assume it doesn't match the
		// specified Booster class.
		return false;
	}
	
	return ((pEventData->noteClass & pBooster->noteClassMask) != 0);
}


/*
* Helper function to add an event/db/agent mapping to the cache.
*/
void AddToCache(string& dbFilename, string& agentName, int eventId, bool beforeEvent, bool global, WORD noteClassMask, string& recordId)
{
	Booster booster(dbFilename, agentName, eventId,  beforeEvent, noteClassMask, recordId);

	DEBUGOUT( "Caching Booster %s in database %s", agentName.c_str(), dbFilename.c_str());

	if (global) {
		gCacheGlobal.AddBooster(booster);
	}
	else {
		BoosterContainer* pContainer = gCacheDbMap.GetContainer(dbFilename, true);

		pContainer->AddBooster(booster);
	}
}

/*
* Helper function to get text out of the summary item buffer.
*/
string GetTextFromSummaryBufferItem(BYTE* pValue, WORD itemLen)
{
	WORD type;

	if (itemLen < sizeof(WORD)) {
		throw runtime_error("Summary buffer value is empty");
	}

	memcpy(&type, pValue, sizeof(WORD));

	if (type != TYPE_TEXT) {
		throw runtime_error("Summary buffer value is not TYPE_TEXT");
	}
	
	return string((char*)(pValue + sizeof(WORD)), itemLen - sizeof(WORD));
}

/*
* Ensure the cache is reasonably up to date.
*/
void EnsureCache(bool firstTime)
{
	STATUS err;
	AUTO_DBHANDLE hConfigDb;

	TIMEDATE lastOpenTime = gConfigLastModified;

	DWORD tickCount = GetTickCount();

	if (!firstTime && tickCount - gLastRefreshTimeTickCount <= gEnsureCacheMs) {
		// not enough time has passed since the last update, just return
		return;
	}

	gLastRefreshTimeTickCount = tickCount; // reset the last refresh time to right now

	if (err = NSFDbOpenExtended(gConfigDbPath, 0, NULLHANDLE, &lastOpenTime, &hConfigDb, &gConfigLastModified, NULL))
	{
		if (err == ERR_NO_MODIFIED_NOTES)
			return; // nothing has changed, the cache is up to date
		else
			CheckStatus(err, "Couldn't open configuration database %s.", gConfigDbPath);
	}

	DEBUGOUT("Loading records into cache");

	// the config db has changes. clear out the cache and load up again from the database

	gCacheDbMap.Clear();
	gCacheGlobal.Clear();

	// open the collection
	
	AUTO_HCOLLECTION hCol;
	NOTEID nidConfigView  = 0;

	CheckStatus( 
		NIFFindView(hConfigDb, CONFIG_VIEW, &nidConfigView) , 
		"Error finding the view " CONFIG_VIEW);

	CheckStatus(
		NIFOpenCollection(
							hConfigDb,
							hConfigDb,
							nidConfigView,
							0,
							NULLHANDLE,
							&hCol,
							NULLHANDLE,
							NULL,
							NULLHANDLE,
							NULLHANDLE),
		"Couldn't open collection");

	char systemName[MAXUSERNAME+1];

	// get the name of the user or the server where we are running
    CheckStatus(
		SECKFMGetUserName(systemName),
		"Couldn't get user or server name.");

	int totalCacheEntriesLoaded;

	totalCacheEntriesLoaded = LoadCacheEntries(hCol, systemName); // get all our entries

	// commented out, because the template doesn't support this yet.
	// totalCacheEntriesLoaded += LoadCacheEntries(hCol, "*"); // get the entries that belong to everyone

	if (gIsMainProcess) {
		if (firstTime) {
			ConsoleOut("%i Booster(s) loaded.", totalCacheEntriesLoaded);
		}
		else {
			ConsoleOut("Configuration changes detected. %i Booster(s) loaded.", totalCacheEntriesLoaded);
		}
	}

}

/*
* Load all the boost records for the given system in the cache and return the number loaded.
*/
int LoadCacheEntries(HCOLLECTION hCol, char* pSystemName)
{
	STATUS error;
	COLLECTIONPOSITION colPos;
	DWORD totalNumberLeft;
	WORD signalFlags;
	int totalNumberAdded = 0;

	error = NIFFindByName (
		              hCol,	    
		              pSystemName,		    
		              FIND_CASE_INSENSITIVE | FIND_ACCENT_INSENSITIVE, 
		              &colPos,		    
		              &totalNumberLeft);

	if (error == ERR_NOT_FOUND) {
		return 0; // no records found, return normally
	}

	do {
		AUTO_HANDLE hBuffer;
		DWORD numberReturned;

		CheckStatus( NIFReadEntries(
							hCol,	
							&colPos,		
							NAVIGATE_CURRENT,
							0,		
							NAVIGATE_NEXT,	   
							totalNumberLeft,
							READ_MASK_SUMMARYVALUES,	
							&hBuffer,
							NULL,		 
							NULL,		 
							&numberReturned,	  
							&signalFlags));

		if (numberReturned && hBuffer)
		{
			totalNumberLeft -= numberReturned;

			// We now have buffer with some number (numberReturned) of entries. Add each
			// entry to the cache.

			// get a pointer to the first item buffer
			HANDLE_LOCK<ITEM_VALUE_TABLE> lockedMem (hBuffer);
			ITEM_VALUE_TABLE* pTable = lockedMem.Lock();

			while (numberReturned--) {				
				try {
					// collect all the data out of the item table and add to cache
					
					if (pTable->Items >= 7)
					{
						WORD* pLength = (WORD*)(pTable + 1);
						BYTE* pValue = (BYTE*)(pLength + pTable->Items);
						
						string dbName;
						string agentName;
						string temp;
						int eventId;
						bool beforeEvent;
						bool global;
						WORD noteClassMask;
						string recordId;

						// the first item is the system name, skip it
						pValue += *pLength++; //advance value and length pointers to next item

						// get the database filename
						dbName = GetTextFromSummaryBufferItem(pValue, *pLength);

						// get the agent name
						pValue += *pLength++; //advance value and length pointers to next item
						agentName = GetTextFromSummaryBufferItem(pValue, *pLength);

						// get the event
						pValue += *pLength++; //advance value and length pointers to next item
						temp = GetTextFromSummaryBufferItem(pValue, *pLength);
						eventId = atoi(temp.c_str());

						// get if its a before of after event
						pValue += *pLength++; //advance value and length pointers to next item
						temp = GetTextFromSummaryBufferItem(pValue, *pLength);
						beforeEvent = (atoi(temp.c_str()) == 1);


						// get if its a global event
						pValue += *pLength++; //advance value and length pointers to next item
						temp = GetTextFromSummaryBufferItem(pValue, *pLength);
						global = (atoi(temp.c_str()) == 1);


						// get the note classes it maps to
						pValue += *pLength++; //advance value and length pointers to next item
						temp = GetTextFromSummaryBufferItem(pValue, *pLength);
						noteClassMask = atoi(temp.c_str());

						// get the record id
						pValue += *pLength++; //advance value and length pointers to next item
						recordId = GetTextFromSummaryBufferItem(pValue, *pLength);

						AddToCache(dbName, agentName, eventId, beforeEvent, global, noteClassMask, recordId);
						totalNumberAdded++;
					}
					else {
						throw runtime_error("Too few columns.");
					}
				}
				catch (exception& e) {
					ConsoleOut("An error occurred loading a record from the configuration database, continuing to next record: ", e.what());
				}
				
				//skip to next item table
				pTable = (ITEM_VALUE_TABLE*)(((BYTE*)pTable) + pTable->Length);
			}
		}

	} while ((signalFlags & SIGNAL_MORE_TO_DO));

	return totalNumberAdded;
}


/*
* Run the named LotusScript agent against an hNote.
*/
void RunAgent(DBHANDLE hDb, const string& agentName, NOTEHANDLE hNote)
{
	char dbName[MAXPATH + 1];
	NOTEID nid;
	AUTO_HAGENT  hAgent;
	AUTO_HAGENTCTX  hAgentCtx;

	
	CheckStatus(
			NSFDbPathGet(hDb, dbName, NULL),
			"Couldn't get db path of handle 0x%x", hDb);

	DEBUGOUT("Attempting to run agent %s in %s", agentName.c_str(), dbName);

	CheckStatus(
		NIFFindDesignNoteExt(hDb, agentName.c_str(), NOTE_CLASS_FILTER, DFLAGPAT_AGENTSLIST, &nid, 0),
		"Couldn't find agent %s in %s", agentName.c_str(), dbName
		);

	CheckStatus(
		AgentOpen(hDb, nid, &hAgent),
		"Couldn't open agent %s in %s", agentName.c_str(), dbName
		);

	CheckStatus(
		AgentCreateRunContext(hAgent,  NULL, AGENT_SECURITY_OFF, &hAgentCtx),
		"Couldn't create context for agent %s in %s", agentName.c_str(), dbName
		);

	CheckStatus(
		AgentSetDocumentContext( hAgentCtx, hNote),
		"Couldn't set document context for agent %s in %s", agentName.c_str(), dbName
		);

	CheckStatus(
		AgentRun(hAgent, hAgentCtx, 0, AGENT_REOPEN_DB),
		"Error running agent %s in %s", agentName.c_str(), dbName
		);

	DEBUGOUT("Finished running agent %s", agentName.c_str());
}

/*
* Get the Event data. Meant to be called from a Lotuscript agent.
* Returns 1 if the event data was returned, returns false if it could
* not be found.
*/
WORD LSGetEventData(LSEventData* pLSEventData)
{
	EventData* pEventData = (EventData*)eventTls.Get();
	if (pEventData) {
		pLSEventData->currentStatus = pEventData->currentStatus;
		pLSEventData->noteId = pEventData->noteId;
		pLSEventData->eventId = pEventData->eventId;
		pLSEventData->isPreEvent = (pEventData->isPreEvent ? 0xFFFF : 0);
		pLSEventData->updateFlags = pEventData->updateFlags;
		pLSEventData->openFlags = pEventData->openFlags;
		
		UNID unid = pEventData->noteUnid;

		sprintf( pLSEventData->noteUnidStr, 
				"%08X%08X%08X%08X", 
				unid.File.Innards[1], unid.File.Innards[0], 
				unid.Note.Innards[1], unid.Note.Innards[0]
				);

		if (pEventData->hDb) {
			NSFDbUserNameGet(pEventData->hDb, pLSEventData->userName, sizeof(pLSEventData->userName));
		}
		else {
			strcpy(pLSEventData->userName, "");
		}
		return 1;
	}
	return 0;
}

/*
* Set the event return status. Meant to be called from LotusScript agent.
*/
void LSSetReturnStatus(STATUS newReturnStatus)
{
	EventData* pEventData = (EventData*)eventTls.Get();
	if (pEventData) {
		pEventData->newStatus = newReturnStatus;
	}
}

/*
* Set the event return Notes status. Meant to be called from LotusScript agent during NSFNoteOpenXXX.
*/
STATUS LSSetReturnNote(NOTEHANDLE hNote)
{
	STATUS err = NOERROR;
	if (hNote == NULLHANDLE) {
		return ERR_HDL_NULL;
	}

	EventData* pEventData = (EventData*)eventTls.Get();
	if (pEventData == NULL) {
		return ERR_UNKNOWN;
	}

	if (pEventData->eventId != EM_NSFNOTEOPENEXTENDED) {
		return ERR_UNKNOWN;
	}

	NOTEHANDLE hNewNote;

	err = NSFNoteCopy(hNote, &hNewNote);

	if (err) {
		return err;
	}

	// make sure the new note has the hDb that the caller of 
	// NSFNoteOpen passed in
	NSFNoteSetInfo(hNewNote, _NOTE_DB, &pEventData->hDb);

	pEventData->hNewNote = hNewNote;

	pEventData->newStatus = NOERROR;

	return NOERROR;
}

bool IsMainProcess()
{
	char buffer[MAX_PATH];
	
	DWORD nameLen = GetModuleFileName( NULL, buffer, sizeof(buffer));

	if (nameLen == 0) {
		// huh? whatevah!
		return false;
	}

	char* pFileName = PathFindFileName(buffer);

	if (_stricmp(pFileName, "nlnotes.exe") == 0) { // main client process
		return true;
	}
	if (_stricmp(pFileName, "nserver.exe") == 0) { // main server process
		return true;
	}

	return false;
}

/*
* Handler for when a memory allocation fails. This way failed allocs are turn
* into exceptions.
*/
void __cdecl ThrowExceptionNewHandler()
{
   throw bad_alloc(); // throw an exception
}
