// Created by Damien Katz and placed into the public domain. Enjoy.

#ifndef CACHE_OBJECTS_DEFS
#define CACHE_OBJECTS_DEFS

/*
* This contains the definitions of the classes used for the cached event to agent mappings.
*/

#define CASE_INSENSITIVE_FILENAMES

#include <vector>

using namespace std;

/*
* Represents an Booster record
*/
class Booster
{
public:
	Booster(string& dbName_in, string& agentName_in, int eventId_in, bool isPreEvent_in, WORD noteClassMask_in, string& recordId_in)
		: dbName(dbName_in), 
		agentName(agentName_in), 
		eventId(eventId_in), 
		isPreEvent(isPreEvent_in), 
		noteClassMask(noteClassMask_in),
		recordId(recordId_in)
	{

#ifdef CASE_INSENSITIVE_FILENAMES
	MakeLowercase(dbName);	// makes file name case insensitive for lookups
							// but only on file systems where case is isn't sensitive
#endif

	MakeLowercase(agentName); // makes agent name case insensitive for lookups
	}

	string dbName;
	string agentName;
	int eventId;
	bool isPreEvent;
	WORD noteClassMask;
	string recordId;
};

/*
* Contains a container of Boosters. Allows for fast lookups.
*/
class BoosterContainer
{
public:
	
	/*
	* Clear out all the Boosters.
	*/
	void Clear()
	{
		preEventBoosterMap.clear();
		postEventBoosterMap.clear();
	}

	/*
	* Add an Booster.
	*/
	void AddBooster(Booster& booster)
	{

		if (booster.isPreEvent) {
			preEventBoosterMap[booster.eventId].push_back(booster);
		}
		else {
			preEventBoosterMap[booster.eventId].push_back(booster);
		}
	}

	/*
	* Get a collection of Boosters that match the event id and the preEvent flag. If preEvent is true
	* then it returns all pre-event Boosters with the specified eventId, other returns all post-event
	* agents for the event id.
	*/
	vector<Booster> GetBoosters(int eventId, bool preEvent)
	{
		map<int, vector<Booster> >& theMap = (preEvent ? preEventBoosterMap : postEventBoosterMap);

		map<int, vector<Booster> >::const_iterator itor;
		itor = theMap.find(eventId);
		if (itor == theMap.end()) {
			return vector<Booster>();
		}
		else {
			return itor->second;
		}
	}

private:
	map<int, vector<Booster> > preEventBoosterMap;
	map<int, vector<Booster> > postEventBoosterMap;
};


/*
* Contains a map of db filenames to BoosterContainers's
*/
class ContainerToFilenameMap
{
public:
	/*
	* Empty the map.
	*/
	void Clear()
	{
		map.clear();
	}

	/*
	* Get the named BoosterContainer out the map. Will return NULL if it doesn't exist, or 
	* if the second optional param is true, will create and insert a new BoosterContainer object 
	* if it doesn't exist already.
	*/
	BoosterContainer* GetContainer(std::string filename, bool createifMissing=false)
	{
#ifdef CASE_INSENSITIVE_FILENAMES
		MakeLowercase(filename); // makes searching case insensitive.
#endif 

		DbMapInternal::iterator itor = map.find(filename);

		if (itor != map.end())
			return &(itor->second);

		if (createifMissing)
		{
			map[filename] = BoosterContainer();
			
			return &(map[filename]);
		}
		else
		{
			return NULL;
		}
	}

private:
	typedef std::map<std::string, BoosterContainer > DbMapInternal;
	DbMapInternal map;

};

#endif // CACHE_OBJECTS_DEFS