// Created by Damien Katz and placed into the public domain. Enjoy.

/******************************************************************************
*
* NotesWrapper.h - Provides classes and helper to make intergrating the Notes CAPI
* into a C++ project easy and lightweight . This file provides exception safe
* automatic wrappers for Notes handles and other Notes resources that must be
* freed. The wrappers have implicit cast operators so the wrappers can be used
* with the C API using the same systax as the native types.
*
* To use each wrapper type, you must make sure to include the corresponding CAPI
* header in your source file.
*
* C API Type:		Wrapper Class:			Also #include:
* ========================================================
* HANDLE			AUTO_HANDLE				osmem.h
* NOTEHANDLE		AUTO_NOTEHANDLE			nsfnote.h
* DBHANDLE			AUTO_DBHANDLE			nsfdb.h
* HAGENT			AUTO_HAGENT				agents.h
* HAGENTCTX			AUTO_HAGENTCTX			agents.h
* HCOLLECTION		AUTO_HCOLLECTION		nif.h
* HEMREGISTRATION	AUTO_HEMREGISTRATION	extmgr.h
*
*
* The AUTO_xxx classes all derive from the same template class  so they all 
* behave identically.
*
* Most of the functionality in the classes is in operator overloads so that 
* they can used just like regular handles. There are additional methods as
* well:
*
* HANDLE Release();
*		Causes the wrapper to return the handle to caller and resets itself to
*		null. DOES NOT FREE THE HANDLE!
*
* void Free();
*		Free the handle and resets self to null.
*
* bool IsNull();
*		Returns true if the internal handle is null.
*
*
* For the most part use the wrapper classes exactly like you would use the 
* corresponding C API type, the only difference is that you don't need to
* explicity free the handle, it frees itself. Instead, if you have something that
* should NOT be freed by the wrapper, then call the wrappers Release() method.
* This will cause to the wrapper to zero the enclosed handle member, but not free
* the actual handle. It's then the caller's responsibilty to free the handle.
*
* The wrappers do not throw exceptions but they can be safely used with or
* without exceptions.
*
* All wrapper classes have the same interface and methods. See the comments
* in the template class NotesWrapper::Handle for more info about the methods.
*
*
************** Locking an AUTO_HANDLE ****************
*
* The HANDLE_LOCK class is used to get to the handle's memory.
*
* For the HANDLE type, the CAPI gives you access to the memory using either
* OSLock(...) or OSLockObject(...) to get the pointer. However, the CAPI requries
* that for every lock call, a corresponding unlock call should be made before
* freeing the memory. The HANDLE_LOCK class automatically takes care of that,
* and ensures the HANDLE is unlocked upon destruction.
* 
* 
*****************  Converting STATUS error to Exceptions  *****************  
*
* Also included is the exception class notes_error (a subclass of std::runtime_error)
* and the CheckStatus() function that makes error handling easier. See the comments
* at the end of this file for more information.
*
*

*****************  Examples  *****************

Below are two example functions that do the same thing, return the most
recently modified of two document. The first is written using the NotesWrapper Api
and exception and the other is using the plain C Api and STATUS's. The purpose is
to show how to use the wrappers

This example uses the NotesWrapper Api. Any error is communicated with an exception.
That the handles opened with NSFNoteOpen are never explicitly freed, instead the 
AUTO_NOTEHANDLE object destructors take care of that.

NOTEHANDLE GetLatestNote(DBHANDLE hDb, NOTEID nid1, NOTEID nid2)
{
	using namespace NotesWrapper;

	AUTO_NOTEHANDLE note1;
	AUTO_NOTEHANDLE note2;
	
	// First open the notes

	CheckStatus(
		NSFNoteOpen(hDb, nid1, 0, &note1)
		);

	CheckStatus(
		NSFNoteOpen(hDb, nid2, 0, &note2)
		);
	
	// Get their modified dates

	TIMEDATE td1;
	TIMEDATE td2;
	
	NSFNoteGetInfo(note1,  _NOTE_MODIFIED, &td1);
	NSFNoteGetInfo(note2,  _NOTE_MODIFIED, &td2);

	// return the one that is most recent

	if (TimeDateCompare(&td1,&td2) > 0) {
		return note1.Release(); // return handle 1 back to the caller
	}
	else {
		return note2.Release(); // return handle 2 back to the caller
	}
}

This example uses the plain C Api. Any error is communicated with a non-zero return status.
STATUS GetLatestNote(DBHANDLE hDb, NOTEID nid1, NOTEID nid2, NOTEHANDLE* pRetHandle)
{
	STATUS err;
	NOTEHANDLE note1;
	NOTEHANDLE note2;
	
	// First open the notes

	if (err = NSFNoteOpen(hDb, nid1, 0, &note1)) {
		return err;
	}

	if (err = NSFNoteOpen(hDb, nid2, 0, &note2)) {
		NSFNoteClose(note1);
		return err;
	}

	// Get their modified dates

	TIMEDATE td1;
	TIMEDATE td2;
	
	NSFNoteGetInfo(note1, _NOTE_MODIFIED, &td1);
	NSFNoteGetInfo(note2, _NOTE_MODIFIED, &td2);

	// return the one that is most recent

	if (TimeDateCompare(&td1,&td2) > 0) {
		NSFNoteClose(note2);
		*pRetHandle = note1; // return handle 1 back to the caller
	}
	else {
		NSFNoteClose(note1);
		*pRetHandle = note2; // return handle 2 back to the caller
	}

	return NOERROR;
}
*/

#ifndef NOTES_WRAPPER_DEFS
#define NOTES_WRAPPER_DEFS

#include <stdexcept>
#include <stdio.h>

#pragma pack(push, 1) // set packing to one byte alignment
	#include "global.h"
	#include "osmisc.h"
#pragma pack(pop) // set back

namespace NotesWrapper
{

	/**
	* The template class for all the handles types.
	*/
	template<class _THandle, typename _TFunc, _TFunc _Func> class Handle
	{		
	public:
		Handle()
			: h(0)
		{
		}

		Handle(_THandle hIn)
			: h(hIn)
		{
		}

		/*
		* If someone passes in a handle wrapper to our constructor we takes its value
		* and then clear out the passed-in wrapper.
		*/
		Handle(Handle& hIn)
			: h(hIn.h)
		{
			hIn.h = 0;
		}

		/*
		* Automatically free the contained handle on destruction.
		*/
		~Handle()
		{			
			Free();
		}

		/*
		* Assignment operator. If another handle is passed in we
		* take its value and assign it to ourself.
		*
		* If we already had a handle contained it will be freed.
		*/

		Handle& operator =(_THandle& hIn)
		{
			if (h != hIn) {
				Free(); // Free our value first
				h = hIn;
			}
			return *this;
		}

		/*
		* Assignment operator. If another handle wrapper is passed
		* in we take its value and assign it to ourself, then clear
		* out the passed in value since we now own it.
		*
		* If we already had a handle contained it will be freed.
		*/
		Handle& operator =(Handle& in)
		{
			if (h != in.h) {
				Free(); // Free our value.
				h = in.h;
				in.h = 0;
			}
			return *this;
		}

		/*
		* Returns the enclosed handle value when a caller implicitu or
		* explicitly casts to the enclosed HANDLE type.
		*
		* Example:
		*	AUTO_HANDLE foo(hHandle);
		*	DWORD dwLen;
		*	OSMemGetSize(foo, &dwLen); // "foo" is cast to HANDLE type
		*/
		operator _THandle()
		{
			return h;
		}

		/*
		* Returns a pointer to the enclosed handle when a caller uses the 
		* address-of operator on the wrapper object.
		*
		* Example:
		*	AUTO_HANDLE foo;
		*	OSMemAlloc(0, 100, &foo); // "&foo" is cast to pointer-to-HANDLE type
		*/
		_THandle* operator &()
		{
			return &h;
		}

		/*
		* This releases control of the handle by the wrapper and returns the enclosed handle.
		* The wrapper no longer will know anything about the handle after this is called.
		* This is useful anytime you want the handle to outlive the wrapper.
		* THIS DOES NOT FREE THE HANDLE!!!
		*/
		_THandle Release() {
			_THandle temp = h;
			h = 0;
			return temp;
		}

		/*
		* Frees the handle if non-null
		*/
		void Free()
		{
			if (h) {
				_Func(h);
				h = 0;
			}
		}

		/*
		* Returns true if the enclosed handle is NULL
		*/
		bool IsNull()
		{
			return h == 0;
		}

	private:

		_THandle h;

	};

// Various freeing function signatures
typedef STATUS (LNPUBLIC * HANDLESTATUSROC)(HANDLE); // STATUS LNPUBLIC Free(HANDLE);
typedef void (LNPUBLIC * VOIDVOIDPROC)(void*); // void LNPUBLIC Free(void*);
typedef STATUS (LNPUBLIC * WORDSTATUSPROC)(WORD); // STATUS LNPUBLIC Free(WORD);
typedef STATUS (LNPUBLIC * DWORDSTATUSPROC)(DWORD); // STATUS LNPUBLIC Free(WORD);


#ifdef MEMORY_DEFS

	/*
	* AUTO_HANDLE. Use in place of HANDLE.
	*/ 
	typedef Handle<HANDLE, HANDLESTATUSROC, OSMemFree> AUTO_HANDLE;

	/*
	* HANDLE_LOCK, used to get the memory pointer of a HANDLE. Unlocks the
	* handle once it goes out of scope.
	*/
	template <typename TMemType=void> class HANDLE_LOCK
	{
	public:
		HANDLE_LOCK(HANDLE handleIn)
			: handle(handleIn), pLockedMem(NULL)
		{
		}

		~HANDLE_LOCK()
		{
			if (pLockedMem) {
				OSUnlock(handle);
			}
		}

		/*
		* Lock the handle and get the pointer to the memory. Can be called
		* multiple times without problem.
		*/
		TMemType* Lock()
		{
			if (!pLockedMem) {
				if (handle == NULLHANDLE) {
					throw runtime_error("Couldn't lock null handle");
				}
				pLockedMem = (TMemType*)OSLockObject(handle);
			}
			return pLockedMem;
		}

	private:
		HANDLE handle;
		TMemType* pLockedMem;
	};

#endif

#ifdef NSF_NOTE_DEFS

	/*
	* AUTO_NOTEHANDLE. Use in place of NOTEHANDLE.
	*/ 
	typedef Handle<NOTEHANDLE, HANDLESTATUSROC, NSFNoteClose> AUTO_NOTEHANDLE;

#endif

#ifdef NSF_DB_DEFS

	/*
	* AUTO_DBHANDLE. Use in place of DBHANDLE.
	*/ 
	typedef Handle<DBHANDLE, HANDLESTATUSROC, NSFDbClose> AUTO_DBHANDLE;

#endif

#ifdef AGENTS_DEFS

	/*
	* AUTO_HAGENT. Use in place of HAGENT.
	*/ 
	typedef Handle<HAGENT, VOIDVOIDPROC, AgentClose> AUTO_HAGENT;

	/*
	* AUTO_HAGENTCTX. Use in place of HAGENTCTX.
	*/ 
	typedef Handle<HAGENTCTX, VOIDVOIDPROC, AgentDestroyRunContext> AUTO_HAGENTCTX;

#endif

#ifdef NIF_DEFS

	/*
	* AUTO_HCOLLECTION. Use in place of HCOLLECTION.
	*/ 
	typedef Handle<HCOLLECTION, WORDSTATUSPROC, NIFCloseCollection> AUTO_HCOLLECTION;

#endif

#ifdef EXMGR_DEFS

	/*
	* AUTO_HEMREGISTRATION. Use in place of HEMREGISTRATION.
	*/ 
	typedef Handle<HEMREGISTRATION, DWORDSTATUSPROC, EMDeregister> AUTO_HEMREGISTRATION;

#endif

	/*
	* An exception class that gets thrown by the CheckStatus function. See the comments for
	* CheckStatus for more information.
	*/
	class notes_error : public std::runtime_error
	{
	public:
		notes_error(STATUS errorIn, const char* pMessage=NULL)
			: runtime_error(pMessage), status(errorIn)
		{}

		STATUS status;
	};

	/*
	* This function will throw an notes_error exception if a non-zero STATUS code
	* is passed in. The second argument is an optional C system format string and the subsequent
	* arguments are the values to use in the format string.
	*/
	static void CheckStatus(STATUS err, const char* pMessageFormat=NULL,...)
	{
		if (err) {
			char finalBuff[200];
			char inputBuff[200];
			char errBuffer[200];

			if (pMessageFormat) {
				va_list args;
				va_start( args, pMessageFormat );
				
				// format the input string and args
				_vsnprintf( inputBuff, sizeof(inputBuff), pMessageFormat, args);

				// load up the error string
				OSLoadString(0, err, errBuffer, sizeof(errBuffer)-1);

				// create the final error string
				_snprintf( finalBuff, sizeof(finalBuff), "%s: %s (0x%x)", inputBuff, errBuffer, err);
			
				throw notes_error(err, finalBuff);
			}
		}
	}
};

#endif // NOTES_WRAPPER_DEFS