// Created by Damien Katz and placed into the public domain. Enjoy.

#ifndef AGENT_BOOST_UTIL
#define AGENT_BOOST_UTIL

/*
* This contains the definitions of various classes and functions
* used by the Agent Boost project
*/

/*
* A mutex.
*/
class Mutex
{
public:
	Mutex()
	{
		InitializeCriticalSection(&cs);
	}

	~Mutex()
	{
		DeleteCriticalSection(&cs);
	}

	/*
	* Locks the mutex
	*/
	void Lock()
	{
		EnterCriticalSection(&cs);
	}

	/*
	* Unlocks the mutex
	*/
	void Unlock()
	{
		LeaveCriticalSection(&cs);
	}

private:
	CRITICAL_SECTION cs;
};

/*
* A stack based lock class for Mutex. Locks on initialization, Unlocks on destruction.
*/
class MutexLock
{
public:
	MutexLock(Mutex& mutex_in)
		: mutex(mutex_in)
	{
		mutex.Lock();
	}

	~MutexLock()
	{
		mutex.Unlock();
	}
private:
	Mutex& mutex;
};


void MakeLowercase(std::string& str)
{
	std::transform(str.begin(), str.end(), str.begin(), tolower);
}

/*
* Dumps the printf style buffer and args to the Notes console and flags as debug.
*/
void DebugOut(const char* format, ...)
{
	/*
	FILE* f;
	f = fopen("c:\\test.txt", "a+");
	if (f) { 
		va_list marker;
		va_start( marker, pString );

		vfprintf(f, pString, marker);
		fprintf(f, "\n");
		fclose(f);
	}*/

	char buffer[256];
	va_list marker;
	va_start( marker, format );

	vsprintf(buffer, format, marker);

	AddInLogMessageText("Debug Agent Boost: %s", NOERROR, buffer);

}


/*
* Dumps the printf style buffer and args to the Notes console.
*/
void ConsoleOut(const char* format,  ...)
{
	char buffer[256];
	va_list marker;
	va_start( marker, format );

	vsprintf(buffer, format, marker);

	AddInLogMessageText("Agent Boost: %s", NOERROR, buffer);
}

/*
* A class that allows simple access to a thread local store. Simply declare in a 
* source file then you can use the Get() and Set() members from inside any function.
*
* Example:
*
* ThreadLocalStorage tls;
*
* void foo()
* {
*	int y=5;
*	tls.Set(&y);
*	bar();
* }
*
* void bar()
* {
*	int* pTemp = (int*)tls.Get();
*	printf("The thread local value is %d", *pTemp);
* }
*
* int main (int argc, char **argv)
* {
*	int x=1;
*	tls.Set(&x);
*	bar();
	foo();
*	return 0;
* }
*
* Output:
* The thread local value is 1
* The thread local value is 5
*/ 
class ThreadLocalStorage
{
public:
	ThreadLocalStorage()
	{
		tls = TlsAlloc();
	}

	~ThreadLocalStorage()
	{
		TlsFree(tls);
	}

	void* Get()
	{
		return TlsGetValue(tls);
	}

	void Set(void* p)
	{
		TlsSetValue(tls, p);
	}

private:
	DWORD tls;
};


#endif // AGENT_BOOST